from .base import AutocompleteBase
from .list import AutocompleteList
from .model import AutocompleteModel
from .choice_list import AutocompleteChoiceList
from .template import AutocompleteTemplate
from .generic import AutocompleteGeneric
from .rest_model import AutocompleteRestModel


class AutocompleteListBase(AutocompleteList, AutocompleteBase):
    pass


class AutocompleteChoiceListBase(AutocompleteChoiceList, AutocompleteBase):
    pass


class AutocompleteModelBase(AutocompleteModel, AutocompleteBase):
    pass


class AutocompleteModelTemplate(AutocompleteModel, AutocompleteTemplate):
    pass


class AutocompleteGenericBase(AutocompleteGeneric, AutocompleteBase):
    pass


class AutocompleteGenericTemplate(AutocompleteGeneric, AutocompleteTemplate):
    pass


class AutocompleteRestModelBase(AutocompleteRestModel, AutocompleteBase):
    pass


class AutocompleteRestModelTemplate(AutocompleteRestModel,
                                    AutocompleteTemplate):
    pass



"""
Checks settings.AUTOCOMPLETE_LIGHT_PROJECT_MODELS, expecting::

     # get_model kwargs         # list of text fields to filter on
    AUTOCOMPLETE_LIGHT_PROJECT_MODELS = (
        (('cities_light', 'city'), ['search_names']),
        (('cities_light', 'region'), ['ascii_name', 'name']),
        (('cities_light', 'country'), ['ascii_name', 'name']),
    )

And registers 'ProjectAutocomplete', an optionnal generic autocomplete that can
third party apps may decide to rely on.

This script is to be imported by autocomplete_light.autodiscover().
"""

from django.conf import settings

import autocomplete_light

try:
    setting = settings.AUTOCOMPLETE_LIGHT_PROJECT_MODELS
except AttributeError:
    pass
else:
    autocomplete = autocomplete_light.AutocompleteGenericBase

    for line in setting:
        model_class = get_model(line[0])
        search_fields = line[1]



from .list import AutocompleteList

__all__ = ('AutocompleteChoiceList',)


class AutocompleteChoiceList(AutocompleteList):
    order_by = lambda cls, choice: unicode(choice[1]).lower()

    def choices_for_values(self):
        values_choices = []

        for choice in self.choices:
            if choice[0] in self.values:
                values_choices.append(choice)

        return self.order_choices(values_choices)[0:self.limit_choices]

    def choices_for_request(self):
        requests_choices = []
        q = self.request.GET.get('q', '')

        for choice in self.choices:
            if q in unicode(choice[0]).lower() + unicode(choice[1]).lower():
                requests_choices.append(choice)

        return self.order_choices(requests_choices)[0:self.limit_choices]

    def choice_value(self, choice):
        return choice[0]

    def choice_label(self, choice):
        return choice[1]



from django.core import urlresolvers

__all__ = ('AutocompleteInterface', 'AutocompleteBase')


class AutocompleteInterface(object):
    """
    This is the minimum to implement in a custom Autocomplete class. It has two
    attributes:

    values
        A list of values which validate_values() and choices_for_values()
        should use.

    request
        A request object which choices_for_request() and autocomplete_html()
        should use.

    An autocomplete proposes "choices". A choice has a "value". When the user
    selects a "choice", then it is converted to a "value".
    """

    def __init__(self, request=None, values=None):
        """
        Set the given request and values as instance attributes, casting values
        to list if necessary.
        """
        self.request = request

        if hasattr(values, '__iter__'):
            self.values = values
        else:
            self.values = [values]

    def autocomplete_html(self):
        """
        Return the HTML autocomplete that should be displayed under the text
        input. Use self.request if set.
        """
        raise NotImplemented()

    def validate_values(self):
        """
        Return True if self.values are all valid.
        """
        raise NotImplemented()

    def choices_for_values(self):
        """
        Return the list of choices corresponding to self.values.
        """
        raise NotImplemented()

    def get_absolute_url(self):
        """
        Return the absolute url for this autocomplete, using
        autocomplete_light_autocomplete url
        """
        return urlresolvers.reverse('autocomplete_light_autocomplete', args=(
            self.__class__.__name__,))


class AutocompleteBase(AutocompleteInterface):
    """
    A basic implementation of AutocompleteInterface that renders HTML and
    should fit most cases. However, it requires to overload
    choices_for_request().
    """
    choice_html_format = u'<div data-value="%s">%s</div>'
    autocomplete_html_format = u'%s'

    def choices_for_request(self):
        """
        Return the list of choices that are available. Uses self.request if
        set. Use self.request if set, may be used by autocomplete_html().
        """
        raise NotImplemented()

    def validate_values(self):
        """
        Return True if all the values are available in choices_for_values().
        """
        return len(self.choices_for_values()) == len(self.values)

    def autocomplete_html(self):
        """
        Simple rendering of the autocomplete.
        """
        html = []

        for choice in self.choices_for_request():
            html.append(self.choice_html(choice))

        return self.autocomplete_html_format % ''.join(html)

    def choice_html(self, choice):
        """
        Return a choice formated according to self.choice_html_format.
        """
        return self.choice_html_format % (
            self.choice_value(choice), self.choice_label(choice))

    def choice_value(self, choice):
        """
        Convert a choice to a value.
        """
        return unicode(choice)

    def choice_label(self, choice):
        """
        Convert a choice to a label.
        """
        return unicode(choice)



from autocomplete.case import *
from autocomplete.generic import AutocompleteGenericMock, AutocompleteGenericTestCase

from django import http
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User, Group, Permission

from autocomplete_light.contrib.generic_m2m import GenericModelForm, \
    GenericModelMultipleChoiceField

from generic_m2m_autocomplete.models import ModelGroup


class FormMock(GenericModelForm):
    related = GenericModelMultipleChoiceField(
        widget=autocomplete_light.MultipleChoiceWidget(
            AutocompleteGenericMock))

    class Meta:
        model = ModelGroup


class AutocompleteGenericM2MTestCase(AutocompleteTestCase):
    autocomplete_mock = AutocompleteGenericMock

    def setUp(self):
        self.setUpAuth()

    def test_form(self):
        tests = (
            {
                'fixture': 'name=foo&related=%s-%s' % (
                    self.group_ctype.pk, self.bluesmen.pk),
                'valid': True,
                'result': [self.bluesmen],
            },
            {
                'fixture': 'name=foo&related=%s-%s&related=%s-%s' % (
                    self.group_ctype.pk, self.rockers.pk,
                    self.user_ctype.pk, self.james.pk),
                'valid': True,
                'result': [self.james, self.rockers],
            },
        )

        instance = None
        for test in tests:
            form = FormMock(http.QueryDict(test['fixture']),
                instance=instance)

            self.assertEqual(form.is_valid(), test['valid'])
            instance = form.save()

            self.assertEqual(instance.related.all().generic_objects(),
                test['result'])



from django import template
from django.utils import safestring

register = template.Library()


@register.filter
def autocomplete_light_data_attributes(attributes, prefix=''):
    html = []

    for key, value in attributes.items():
        html.append('data-%s%s="%s"' % (prefix, key.replace('_', '-'), value))

    return safestring.mark_safe(u' '.join(html))


@register.filter
def autocomplete_light_choice_html(choice, autocomplete):
    """Return autocomplete.choice_html(choice)"""
    return safestring.mark_safe(autocomplete.choice_html(choice))



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

from autocomplete_light.generic import GenericModelChoiceField

from .model import AutocompleteModel

__all__ = ['AutocompleteGeneric']


class AutocompleteGeneric(AutocompleteModel):
    choices = None
    search_fields = None

    def choice_value(self, choice):
        """
        Rely on GenericModelChoiceField to return a string containing the
        content type id and object id of the result.

        Because this autocomplete is made for that field, and to avoid code
        duplication.
        """
        field = GenericModelChoiceField()
        return field.prepare_value(choice)

    def validate_values(self):
        """
        Ensure that every choice is part of a queryset.
        """
        assert self.choices, 'autocomplete.choices should be a queryset list'

        for value in self.values:
            if not isinstance(value, str):
                return False

            try:
                content_type_id, object_id = value.split('-')
            except ValueError:
                return False

            try:
                content_type = ContentType.objects.get_for_id(content_type_id)
            except ContentType.DoesNotExist:
                return False

            model_class = content_type.model_class()

            found = False
            for queryset in self.choices:
                if queryset.model != model_class:
                    continue

                if queryset.filter(pk=object_id).count() == 1:
                    found = True
                else:
                    return False

            if not found:
                # maybe a user would cheat by using a forbidden ctype id !
                return False

        return True

    def choices_for_request(self):
        """
        Propose local results and fill the autocomplete with remote
        suggestions.
        """
        assert self.choices, 'autocomplete.choices should be a queryset list'

        q = self.request.GET.get('q', '')

        request_choices = []
        querysets_left = len(self.choices)

        i = 0
        for queryset in self.choices:
            conditions = Q()

            if q:
                for search_field in self.search_fields[i]:
                    conditions |= Q(**{search_field + '__icontains': q})

            limit = ((self.limit_choices - len(request_choices)) /
                querysets_left)
            for choice in queryset.filter(conditions)[:limit]:
                request_choices.append(choice)

            querysets_left -= 1
            i += 1

        return request_choices

    def choices_for_values(self):
        """
        Values which are not found in the querysets are ignored.
        """
        values_choices = []

        for queryset in self.choices:
            ctype = ContentType.objects.get_for_model(queryset.model).pk

            try:
                ids = [x.split('-')[1] for x in self.values
                    if x is not None and int(x.split('-')[0]) == ctype]
            except ValueError:
                continue

            for choice in queryset.filter(pk__in=ids):
                values_choices.append(choice)

        return values_choices



__all__ = ('AutocompleteList',)


class AutocompleteList(object):
    limit_choices = 20
    order_by = lambda cls, choice: unicode(choice).lower()

    def choices_for_values(self):
        values_choices = []

        for choice in self.choices:
            if choice in self.values:
                values_choices.append(choice)

        return self.order_choices(values_choices)[0:self.limit_choices]

    def choices_for_request(self):
        assert self.choices, 'autocomplete.choices is not set'

        requests_choices = []
        q = self.request.GET.get('q', '')

        for choice in self.choices:
            if q in unicode(choice).lower():
                requests_choices.append(choice)

        return self.order_choices(requests_choices)[0:self.limit_choices]

    def order_choices(self, choices):
        return sorted(choices, key=self.order_by)



import types

from django.template import loader

from .base import AutocompleteBase


class AutocompleteTemplate(AutocompleteBase):
    choice_template = None
    autocomplete_template = None

    def get_base_context(self):
        return {
            'request': self.request,
            'autocomplete': self,
        }

    def render_template_context(self, template, extra_context=None):
        context = self.get_base_context()
        context.update(extra_context or {})
        return loader.render_to_string(template, context)

    def autocomplete_html(self):
        if self.autocomplete_template:
            choices = self.choices_for_request()

            return self.render_template_context(self.autocomplete_template,
                {'choices': choices})
        else:
            return super(AutocompleteTemplate, self).autocomplete_html()

    def choice_html(self, choice):
        if self.choice_template:
            return self.render_template_context(self.choice_template,
                {'choice': choice})
        else:
            return super(AutocompleteTemplate, self).choice_html(choice)
